/**
 * Copyright (C) 2003-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */

package com.funambol.framework.core.dm.bootstrap;

import java.io.Serializable;

import java.util.ArrayList;
import java.util.List;
import java.util.Iterator;

/**
 * Corresponds to the &lt;characteristic&gt; tag in a Wap provisioning doc.
 *
 * @version $Id: Characteristic.java,v 1.2 2007/01/11 19:30:39 nichele Exp $
 */
public final class Characteristic implements Serializable {

    private static final long serialVersionUID = 4113564488529304650L;

    // --------------------------------------------------------------- Constants
    public static final String TYPE_PXLOGICAL          = "PXLOGICAL";
    public static final String TYPE_PXPHYSICAL         = "PXPHYSICAL";
    public static final String TYPE_PXAUTHINFO         = "PXAUTHINFO";
    public static final String TYPE_PORT               = "PORT";
    public static final String TYPE_NAPDEF             = "NAPDEF";
    public static final String TYPE_NAPAUTHINFO        = "NAPAUTHINFO";
    public static final String TYPE_VALIDITY           = "VALIDITY";
    public static final String TYPE_BOOTSTRAP          = "BOOTSTRAP";
    public static final String TYPE_CLIENTIDENTITY     = "CLIENTIDENTITY";
    public static final String TYPE_VENDORCONFIG       = "VENDORCONFIG";
    public static final String TYPE_APPLICATION        = "APPLICATION";
    public static final String TYPE_APPADDR            = "APPADDR";
    public static final String TYPE_APPAUTH            = "APPAUTH";
    public static final String TYPE_RESOURCE           = "RESOURCE";
    public static final String TYPE_ACCESS             = "ACCESS";


    // -------------------------------------------------------------- Properties

    private ArrayList children;

    private String type;

    // ------------------------------------------------------------ Constructors

    public Characteristic() {
        children = new ArrayList();
    }

    public Characteristic(String type) {
        this();
        this.type = type;
    }


    /**
     * Adds new <code>Characteristic</code> to this characteristic
     * @param characteristicToAdd the characteristic to add
     */
    public void add(Characteristic characteristicToAdd) {
        children.add(characteristicToAdd);
    }

    /**
     * Adds new <code>Parm</code> to this characteristic
     * @param parmToAdd the parm to add
     */
    public void add(Parm parmToAdd) {
        children.add(parmToAdd);
    }

    /**
     * Gets the type
     * @return the type of this characteristic
     */
    public String getType() {
        return type;
    }

    /**
     * Sets the type of this characteristic
     * @param type the type to sets
     */
    public void setType(String type) {
        this.type = type;
    }

    /**
     * Returns the children list
     * @return the children
     */
    public ArrayList getChildren() {
        return children;
    }


    /**
     * Removes for the children list the parm with the given name
     * @param name String
     */
    public void removeParm(String name) {
        Iterator itChildren = null;
        Object   child      = null;
        Parm     parm       = null;
        if (children != null) {
            itChildren = children.iterator();
            while (itChildren.hasNext()) {
                child = itChildren.next();
                if (child instanceof Parm) {
                    parm = (Parm)child;
                    if (name.equalsIgnoreCase(parm.getName())) {
                        children.remove(parm);
                        return;
                    }
                }
            }
        }
        return ;
    }

    /**
     * Removes for the children list the characteristics with the given type
     * (Maybe there are more characteristics with the same name)
     * @param type String
     */
    public void removeCharacteristic(String type) {
        Iterator       itChildren   = null;
        Object         child        = null;
        Characteristic c            = null;
        List           charToRemove = new ArrayList();
        if (children != null) {
            itChildren = children.iterator();
            while (itChildren.hasNext()) {
                child = itChildren.next();
                if (child instanceof Characteristic) {
                    c = (Characteristic)child;
                    if (type.equalsIgnoreCase(c.getType())) {
                        charToRemove.add(c);
                    }
                }
            }
        }
        children.removeAll(charToRemove);
        return ;
    }

    /**
     * Returns the parm with the given name searching it in the children list.
     * @param name the name to search
     * @return the parm with the given name searching it in the children list.
     *         <code>null</code> if a parm with the specified name doesn't exist.
     */
    public Parm getParm(String name) {
        Iterator itChildren = null;
        Object   child      = null;
        Parm     parm       = null;
        if (children != null) {
            itChildren = children.iterator();
            while (itChildren.hasNext()) {
                child = itChildren.next();
                if (child instanceof Parm) {
                    parm = (Parm)child;
                    if (name.equalsIgnoreCase(parm.getName())) {
                        return parm;
                    }
                }
            }
        }
        return null;
    }

    /**
     * Sets a parm with the given name and value.
     * If a parm with the given name is found in the children list, his value is
     * overwritten, otherwise a new Parm is created and add at the given position
     * @param name String
     * @param value String
     * @param position the position of the parm.
     */
    public int setParm(String name, String value, int position) {
        Iterator itChildren = null;
        Object   child      = null;
        Parm     parm       = null;
        int      pos        = 0;
        if (children != null) {
            itChildren = children.iterator();
            while (itChildren.hasNext()) {
                child = itChildren.next();
                if (child instanceof Parm) {
                    parm = (Parm)child;
                    if (name.equalsIgnoreCase(parm.getName())) {
                        parm.setValue(value);
                        return pos;
                    }
                }
                pos++;
            }
        }
        parm = new Parm(name, value);
        children.add(position, parm);
        return position;
    }

    /**
     * Returns the characteristic with the given type searching it in the children list.
     * @param type the type to search
     * @return the characteristic with the given type searching it in the children list.
     *         <code>null</code> if a characteristic with the specified type doesn't exist.
     */
    public Characteristic getCharacteristic(String type) {
        Iterator itChildren = null;
        Object   child      = null;
        Characteristic characteristic = null;

        if (children != null) {
            itChildren = children.iterator();
            while (itChildren.hasNext()) {
                child = itChildren.next();
                if (child instanceof Characteristic) {
                    characteristic = (Characteristic)child;
                    if (type.equalsIgnoreCase(characteristic.getType())) {
                        return characteristic;
                    }
                }
            }
        }
        return null;
    }
}
