/**
 * Copyright (C) 2006-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.framework.core.dm.ddf;

import java.io.Serializable;
import java.util.Arrays;
import java.util.ArrayList;
import java.util.Iterator;

import org.apache.commons.lang.builder.ToStringBuilder;

import com.funambol.framework.core.*;

/**
 * Represents a MgmtTree in DM Tree 1.2. The Management Tree is a hierarchical arrangement 
 * of managed objects in a device, which defines the management view of the device. 
 * 
 * @version $Id: MgmtTree.java,v 1.2 2007/01/11 19:30:39 nichele Exp $
 */
public class MgmtTree implements Serializable {

    // -------------------------------------------------------------- Properties
    private ArrayList treeNodes = new ArrayList();
    private VerDTD verDTD;
    private String man;
    private String mod;
    
    // ------------------------------------------------------------- Constructor
    public MgmtTree() {
    }

    /**
     * Creates a new MgmtTree
     * @param treeNodes     the list of children nodes
     * @param verDTD        the DTD version
     * @param man           the manufacter identifier
     * @param mod           the model identifier (manufacter specified string)
     */
    public MgmtTree(Node[] treeNodes,
                    VerDTD verDTD   ,
                    String man      ,
                    String mod      ){
        
        if (treeNodes == null) {
            treeNodes = new Node[0];
        }
        setTreeNodes(treeNodes);
        setVerDTD(verDTD);
        this.man = man;
        this.mod = mod;
    }
    
    // ---------------------------------------------------------- Public Methods

    /**
     * Returns the manufacter identifier
     *
     * @return the man property
     */
    public String getMan() {
        return man;
    }

    /**
     * Set the manufacter identifier
     * 
     * @param man the new manufacter identifier
     */
    public void setMan(String man) {
        this.man = man;
    }

    /**
     * Returns the model identifier
     *
     * @return the mod property
     */
    public String getMod() {
        return mod;
    }

    /**
     * Set the model identifier
     * 
     * @param mod the new model identifier
     */
    public void setMod(String mod) {
        this.mod = mod;
    }

    /**
     * Returns the list of nodes
     *
     * @return the treeNodes list
     */
    public ArrayList getTreeNodes() {
        return this.treeNodes;
    }
  
    /**
     * Appends all of the nodes in the array to the end of the nodes collection
     * 
     * @param nodes the new nodes
     */
    public void addTreeNodes(Node[] nodes) {
        if (nodes != null) {
            this.treeNodes.addAll(Arrays.asList(nodes));
        }
    }
    
    /**
     * Appends the new node to the end of the nodes collection
     *  
     * @param node the new node
     */
    public void addNode(Node node) {
        this.treeNodes.add(node);
    }

    /**
     * Appends all of the nodes in the arrayList to the end of the nodes collection
     * 
     * @param nodes the collection of new nodes
     * @throws IllegalArgumentException if the arraylist includes not valid objects 
     */
    public void addTreeNodes(ArrayList nodes) {
        if (nodes == null){
            return;
        }      
        Iterator i = nodes.iterator();
        while (i.hasNext()){
            if (!(i.next() instanceof Node)) {
                throw new IllegalArgumentException("The nodes in the array list are not valid");
            }
        }
        this.treeNodes.addAll(nodes);
    }

    /**
     * Replaces the nodes collection with the new nodes
     * 
     * @param nodes the collection of new nodes
     * @throws IllegalArgumentException if the arraylist includes not valid objects  
     */
    public void setTreeNodes(ArrayList nodes) {
        if (nodes != null) {                 
            Iterator i = nodes.iterator();
            while (i.hasNext()){
                if (!(i.next() instanceof Node)) {
                    throw new IllegalArgumentException("The nodes in the array list are not valid");
                }
            }
            this.treeNodes = nodes;
        } else {
            this.treeNodes.clear();
        }
    }   
    
    /**
     * Replaces the nodes in the collection with the new nodes
     * 
     * @param nodes the new nodes
     */
    public void setTreeNodes(Node[] nodes) {
        this.treeNodes.clear();
        if (nodes != null) {
            this.treeNodes.addAll(Arrays.asList(nodes));
        }
    }

    /**
     * Returns the DTD version
     *
     * @return the verDTD property
     */
    public VerDTD getVerDTD() {
        return verDTD;
    }

    /**
     * Set the DTD version. Specifies the major and minor version identifier of the OMA DM 
     * Description Framework specification used to represent the OMA DM description.
     *
     * @param verDTD the new DTD version
     */    
    public void setVerDTD(VerDTD verDTD) {
        this.verDTD = verDTD;
    }

    /**
     * Returns a String representation of this Management Tree 
     * 
     * @return String
     */
    public String toString() {
        ToStringBuilder sb = new ToStringBuilder(this);
    
        sb.append("treeNodes", treeNodes);
        sb.append("VerDTD", (verDTD == null) ? null : verDTD.getValue());        
        sb.append("man"   , man); 
        sb.append("mod"   , mod);

        return sb.toString();
    }
}
