/**
 * Copyright (C) 2006-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.framework.engine.dm;

import org.apache.commons.lang.builder.ToStringBuilder;

import com.funambol.framework.core.Constants;
import com.funambol.framework.core.VerProto;
import com.funambol.framework.core.dm.ddf.DevInfo;
import com.funambol.framework.security.Sync4jPrincipal;

/**
 * Contains the session context used in the DM process. The SessionContext has:
 * <ul>
 * <li>protocol version</li>
 * <li>sessionId is the content of the SessionID element of the OTA DM message</li>
 * <li>type is the management session type (as specified in the message Alert)</li>
 * <li>info is the device info of the device under management</li> 
 * <li>dmstate is the device management state, which represents a row of the fnbl_dm_state table</li>
 * </ul>
 * 
 */
public class SessionContext implements java.io.Serializable {

    // -------------------------------------------------------------- Constants
    
       public static final VerProto SYNCML_DM_1_1   = Constants.SYNCML_DM_1_1 ;
       public static final VerProto SYNCML_DM_1_2   = Constants.SYNCML_DM_1_2 ;
    
    // -------------------------------------------------------------- Properties

    private String          sessionId          = null;
    private Sync4jPrincipal principal          = null;
    private int             type;
    private DevInfo         devInfo            = null;
    private DeviceDMState   dmstate            = null; 
    private VerProto        verPro             = null;
    
    // ------------------------------------------------------------- Constructor

    /**
     * Creates a new SyncContext
     * @param sessionId     the content of the SessionID element of the DM message
     * @param principal     the principal
     * @param type          the management session type
     * @param devInfo       the device info of the device under management
     * @param dmstate       the device management state
     * @param verPro        the version protocol
     * 
     */
    public SessionContext(String sessionId        ,
                       Sync4jPrincipal principal  ,
                       int             type       ,
                       DevInfo         devInfo    ,
                       DeviceDMState   dmstate    ,
                       VerProto        verPro   ) {

        this.sessionId          = sessionId;
        this.principal          = principal;
        this.type               = type     ;
        this.devInfo            = devInfo  ;
        this.dmstate            = dmstate  ;
        this.verPro             = verPro   ;
        
    }
    
    // ---------------------------------------------------------- Public Methods            

    /**
     * Returns a String representation of this Session Context
     * @return String
     */
    public String toString() {
        ToStringBuilder sb = new ToStringBuilder(this).append("\r\n");

        sb.append("sessionId", sessionId);
        sb.append("principal", principal);
        sb.append("type"     , type);
        sb.append("devInfo"  , devInfo.getDevId());
        sb.append("dmstate"  , dmstate);
        sb.append("verPro"   , verPro.getVersion());
        
        return sb.toString();
    }
    
    /**
     * Returns the principal
     * @return the principal property 
     */
    public Sync4jPrincipal getPrincipal() {
        return principal;
    }

    /**
     * Returns the device info of the device under management
     * @return the devinfo property 
     */    
    public DevInfo getDevInfo() {
        return this.devInfo;
    }

    /**
     * Returns the device management state
     * @return the dmstate property 
     */    
    public DeviceDMState getDmstate() {
        return this.dmstate;
    }
    
    /**
     * Returns the content of the SessionID element of the DM message
     * @return the sessionId property 
     */
    public String getSessionId() {
        return this.sessionId;
    }

    /**
     * Returns the management session type
     * @return the type property 
     */    
    public int getType() {
        return this.type;
    }

    /**
     * Returns the version protocol
     * @return the verPro property 
     */    
    public VerProto getVerPro() {
        return this.verPro;
    }
    
}