/**
 * Copyright (C) 2003-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */

package com.funambol.framework.engine.pipeline;

import java.io.Serializable;
import java.util.logging.Logger;
import java.util.logging.Level;

import com.funambol.framework.logging.FunambolLogger;

import com.funambol.framework.core.SyncML;
import com.funambol.framework.core.Sync4jException;
import com.funambol.framework.core.Util;

/**
 * This class represents the manager of the pipeline. It supplies a hook for adding
 * additional processing and manipulation of the message between the server and the
 * client.
 */
public class PipelineManager implements InputMessageProcessor, OutputMessageProcessor, Serializable {

    // --------------------------------------------------------------- Constants

    // ------------------------------------------------------------ Private data

    private static final Logger log = FunambolLogger.getLogger("engine.pipeline");

    // ------------------------------------------------------------ Constructors

    // -------------------------------------------------------------- Properties

    private InputMessageProcessor  inputProcessors[]  = null;
    private OutputMessageProcessor outputProcessors[] = null;

    /** Getter for property inputProcessors.
     * @return Value of property inputProcessors.
     */
    public InputMessageProcessor[] getInputProcessors() {
        return this.inputProcessors;
    }

    /** Setter for property inputProcessors.
     * @param inputProcessors New value of property inputProcessors.
     */
    public void setInputProcessors(InputMessageProcessor[] inputProcessors) {
        this.inputProcessors = inputProcessors;
    }

    /** Getter for property outputProcessors.
     * @return Value of property outputProcessors.
     */
    public OutputMessageProcessor[] getOutputProcessors() {
        return this.outputProcessors;
    }

    /** Setter for property outputProcessors
     * @param outputProcessors New value of property outputProcessors.
     */
    public void setOutputProcessors(OutputMessageProcessor[] outputProcessors) {
        this.outputProcessors = outputProcessors;
    }

    // ---------------------------------------------------------- Public methods

	public PipelineManager() {
    }

    /** Process the message with the input processors.
     * @param processingContext message processing context
     * @param message the message to be processed
     */
    public void preProcessMessage(MessageProcessingContext processingContext, SyncML message) {
        if (log.isLoggable(Level.FINEST)) {
            log.finest("Starting preprocessing");
        }

        //cycle on inputMessageProcessor[] elements
        int size = inputProcessors.length;
        for (int i=0; i<size; i++) {
            try {
                inputProcessors[i].preProcessMessage(processingContext, message);
            } catch(Sync4jException e) {
                log.info("preProcessMessage error: " + e);
            }
        }
        if (log.isLoggable(Level.FINEST)) {
            log.finest("Preprocessed message: " + Util.toXML(message));
        }
    }

    /** Process the message with the output processors.
     * @param processingContext message processing context
     * @param message the message to be processed
     */
    public void postProcessMessage(MessageProcessingContext processingContext, SyncML message) {
        if (log.isLoggable(Level.FINEST)) {
            log.finest("Starting postprocessing");
            log.finest("Returning message to process: " + Util.toXML(message));
        }

        //cycle on outputMessageProcessor[] elements
        int size = outputProcessors.length;
        for (int i = 0; i < size; i++) {
            try {
                outputProcessors[i].postProcessMessage(processingContext, message);
            } catch (Sync4jException e) {
                log.info("postProcessMessage error: " + e);
            }
        }
    }
}
