/**
 * Copyright (C) 2003-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */

package com.funambol.framework.notification;

import java.security.MessageDigest;
import com.funambol.framework.tools.Base64;
import java.security.NoSuchAlgorithmException;


/**
 * Represents the digest of the Notification message in according with the SyncML DM specification.
 * <p>
 * Compute the digest message in the following form:
 * <p>
 *      <ui>Digest = H(B64(H(serverId:pw)):nonce:B64(H(trigger)))</ui>
 * <p>
 * using MD5 algorithm.
 *
 *
 *
 * @version $Id: DigestNotificationMessage.java,v 1.4 2007/01/11 19:30:40 nichele Exp $
 */
public class DigestNotificationMessage {

    // --------------------------------------------------------------- Constants
    private static final String ALGORITHM_NAME = "MD5";
    private static final String SEPARATOR      = ":";

    // -------------------------------------------------------------- Properties
    private String serverId;
    private String serverPw;
    private byte[] nonce;

    // ------------------------------------------------------------ Constructors

    public DigestNotificationMessage() {}

    /**
     * Creates a new DigestNotificationMessage object with the given parameters
     *
     * @param serverId
     * @param serverPw
     * @param nonce
     */
    public DigestNotificationMessage(final String serverId,
                                     final String serverPw,
                                     final byte[] nonce) {
        setServerId(serverId);
        setServerPw(serverPw);
        setNonce   (nonce   );
    }

    // ---------------------------------------------------------- Public methods
    /**
     * Gets the server identifier
     *
     * @return  the server identifier
     */
    public String getServerId() {
        return serverId;
    }

    /**
     * Sets the server identifier
     *
     * @param  serverId the server identifier
     */
    public void setServerId(String serverId) {
        this.serverId = serverId;
    }

    /**
     * Gets the server password
     *
     * @return  the server password
     */
    public String getServerPw() {
        return serverPw;
    }

    /**
     * Sets the server password
     *
     * @param serverPw the server password
     */
    public void setServerPw(String serverPw) {
        this.serverPw = serverPw;
    }

    /**
     * Gets the nonce
     *
     * @return  the nonce
     */
    public byte[] getNonce() {
        return nonce;
    }

    /**
     * Sets the nonce
     *
     * @param  nonce the nonce; if null, the property nonce is set to an empty array
     */
    public void setNonce(byte[] nonce) {
        this.nonce = (nonce == null) ? new byte[0] : nonce;
    }

    /**
     * Compute the digest for the triggerMessage.
     * <p>
     * The digest message is calculate as:
     * <p>
     * Digest = H(B64(H(serverId:pw)):nonce:B64(H(trigger)))
     * @param triggerMessage the message of which calculate the digest
     * @throws NoSuchAlgorithmException if MD5 algorithm is not available
     * @return byte[] the digest messge
     */
    public byte[] computeDigestMessage(byte[] triggerMessage) throws NoSuchAlgorithmException {

        String cred = serverId + SEPARATOR + serverPw;

        MessageDigest md =  MessageDigest.getInstance(ALGORITHM_NAME);

        byte[] digestTriggerMessage = null;
        byte[] b64DigestTriggerMessage = null;

        byte[] digestCred = null;
        byte[] b64DigestCred = null;
        byte[] digest = null;

        // H(trigger)
        digestTriggerMessage = md.digest(triggerMessage);

        // B64(H(trigger))
        b64DigestTriggerMessage = Base64.encode(digestTriggerMessage);

        md.reset();

        // H(serverId:pw)
        digestCred = md.digest(cred.getBytes());

        // B64(H(serverId:pw))
        b64DigestCred = Base64.encode(digestCred);

        md.reset();

        // Creates a unique buffer containing the bytes to digest
        //
        byte[] buf = new byte[b64DigestCred.length + 2 + nonce.length + b64DigestTriggerMessage.length];

        System.arraycopy(b64DigestCred, 0, buf, 0, b64DigestCred.length);
        buf[b64DigestCred.length] = (byte)':';
        System.arraycopy(nonce, 0, buf, b64DigestCred.length+1, nonce.length);
        buf[b64DigestCred.length + nonce.length + 1] = (byte)':';
        System.arraycopy(b64DigestTriggerMessage, 0, buf, b64DigestCred.length + nonce.length + 2, b64DigestTriggerMessage.length);

        digest = md.digest(buf);

        return digest;
    }

}
