/**
 * Copyright (C) 2006-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.server.bootstrap;

import java.io.ByteArrayOutputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.jibx.runtime.*;

import com.funambol.framework.core.*;
import com.funambol.framework.core.dm.bootstrap.BasicAccountInfo;

import com.funambol.framework.engine.dm.TreeNode;
import com.funambol.framework.engine.dm.Util;

import com.funambol.framework.logging.FunambolLogger;
import com.funambol.framework.logging.FunambolLoggerName;

import com.funambol.framework.notification.NotificationException;

import com.funambol.framework.tools.Base64;
import com.funambol.framework.tools.WBXMLTools;


/**
 * This is the builder to create a bootstrap message according to the PLAIN profile
 * @version $Id: PlainBootstrapMessageBuilder.java,v 1.2 2007/01/11 19:30:50 nichele Exp $
 */
public class PlainBootstrapMessageBuilder {

    // --------------------------------------------------------------- Constants

    //Settings
    public static final String ACC_NAME             = "/Name";
    public static final String ACC_AUTH_PREF        = "/AuthPref";
    //Server
    public static final String ACC_SERVER_ADDR      = "/Addr";
    public static final String ACC_SERVER_PORT_NBR  = "/PortNbr";
    public static final String ACC_SERVER_ADDR_TYPE = "/AddrType";
    public static final String ACC_SERVER_ID        = "/ServerId";
    public static final String ACC_SERVER_PASSWORD  = "/ServerPW";
    public static final String ACC_SERVER_NONCE     = "/ServerNonce";
    //Client
    public static final String ACC_CLIENT_USERNAME  = "/UserName";
    public static final String ACC_CLIENT_PASSWORD  = "/ClientPW";
    public static final String ACC_CLIENT_NONCE     = "/ClientNonce";

    // -------------------------------------------------------------- Properties

    // ------------------------------------------------------------ Private data

    private Logger log = null;

    // ------------------------------------------------------------- Constructor

    /**
     * Creates a new PlainBootstrapMessageBuilder
     */
    public PlainBootstrapMessageBuilder() {
        log = FunambolLogger.getLogger(FunambolLoggerName.SERVER_DM_BOOTSTRAP);
    }

    // ---------------------------------------------------------- Public methods

    /**
     * Creates the plain bootstrap message using the given information
     * @param basicAccountInfo the basic account info
     * @param nodes the nodes to set in the plain bootstrap message
     * @throws NotificationException if an error occurs
     * @return the plain bootstrap message (encoded in WBXML)
     */
    public byte[] buildMessage(BasicAccountInfo basicAccountInfo,
                               TreeNode[]       nodes)
    throws NotificationException {

        if (log.isLoggable(Level.FINEST)) {
            log.finest("Calling buildMessage in PlainBootstrapMessageBuilder");
        }

        Map bootstrapNodes = accountInfoNodes(basicAccountInfo);

        if (nodes != null) {
            for (int i = 0; i < nodes.length; i++) {
                if (!bootstrapNodes.containsKey(nodes[i].getName())) {
                    bootstrapNodes.put(nodes[i].getName(), nodes[i]);
                }
            }
        }

        SyncML syncMLPlainBootstrapMessage = getSyncMLBootstrapMessage(bootstrapNodes, basicAccountInfo);

        byte[] message = null;

        try {

            ByteArrayOutputStream bout = new ByteArrayOutputStream();

            IBindingFactory f = BindingDirectory.getFactory(SyncML.class);
            IMarshallingContext c = f.createMarshallingContext();
            c.setIndent(0);
            c.marshalDocument(syncMLPlainBootstrapMessage, "UTF-8", null, bout);

            String inputXml = new String(bout.toByteArray());
            inputXml = metInfNamespaceHandler(inputXml);

            if (log.isLoggable(Level.FINEST)) {
                log.finest("PlainBootstrapMessage - xml: " + inputXml);
            }

            message = WBXMLTools.toWBXML(inputXml, Constants.DTD_1_1.getValue());

        } catch (Exception ex) {
            throw new NotificationException("Error during xml to wbxml transformation", ex);
    }

    return message;
}

    /**
     * This is a solution in order to obviate to a JiBX bug: it does
     * not allow to declare the namespace to level of structure.
     *
     * @param msg the server response
     *
     * @return the response with namespace correctly added into MetInf element
     */
    private String metInfNamespaceHandler(String msg) {
        int s = 0;
        int e = 0;

        StringBuffer response = new StringBuffer();
        while (( e = msg.indexOf("<Meta", s)) >= 0) {

            response = response.append(msg.substring(s, e));

            int a = msg.indexOf("</Meta>", e);
            String meta = msg.substring(e, a);

            meta = meta.replaceAll("<Type>"   , "<Type xmlns='syncml:metinf'>");
            meta = meta.replaceAll("<Format>" , "<Format xmlns='syncml:metinf'>");
            meta = meta.replaceAll("<Mark>"   , "<Mark xmlns='syncml:metinf'>");
            meta = meta.replaceAll("<Size>"   , "<Size xmlns='syncml:metinf'>");
            meta = meta.replaceAll("<Anchor>" , "<Anchor xmlns='syncml:metinf'>");
            meta = meta.replaceAll("<Version>", "<Version xmlns='syncml:metinf'>");
            meta = meta.replaceAll("<NextNonce>" , "<NextNonce xmlns='syncml:metinf'>");
            meta = meta.replaceAll("<MaxMsgSize>", "<MaxMsgSize xmlns='syncml:metinf'>");
            meta = meta.replaceAll("<MaxObjSize>", "<MaxObjSize xmlns='syncml:metinf'>");
            meta = meta.replaceAll("<EMI>"    , "<EMI xmlns='syncml:metinf'>");
            meta = meta.replaceAll("<Mem>"    , "<Mem xmlns='syncml:metinf'>");

            s = a + 7;
            response.append(meta).append("</Meta>");
        }
        return response.append(msg.substring(s, msg.length())).toString();
    }


    /**
     * Returns a linked hash map of the bootstrap account properties
     *
     * @param basicAccountInfo
     * @return the linked hash map of the basic account info
     */
    private Map accountInfoNodes(BasicAccountInfo basicAccountInfo) {

        Map accountNodes = new LinkedHashMap();
        String accountURI = "./SyncML/DMAcc/" + basicAccountInfo.getAccountName();

        //Settings

        accountNodes.put(accountURI, new TreeNode(accountURI, null, TreeNode.FORMAT_NODE));

        accountNodes.put(accountURI + ACC_AUTH_PREF,
                         new TreeNode(accountURI + ACC_AUTH_PREF, basicAccountInfo.getAuthPref()));

        //Server
        accountNodes.put(accountURI + ACC_SERVER_ADDR,
                         new TreeNode(accountURI + ACC_SERVER_ADDR, basicAccountInfo.getServerAddress()));

        accountNodes.put(accountURI + ACC_SERVER_PORT_NBR,
                         new TreeNode(accountURI + ACC_SERVER_PORT_NBR,
                                      Integer.toString(basicAccountInfo.getServerPort()),
                                      TreeNode.FORMAT_INT)
                         );

        accountNodes.put(accountURI + ACC_SERVER_ADDR_TYPE,
                         new TreeNode(accountURI + ACC_SERVER_ADDR_TYPE,
                                      Integer.toString(basicAccountInfo.getAddressType()),
                                      TreeNode.FORMAT_INT)
                         );

        accountNodes.put(accountURI + ACC_SERVER_ID,
                         new TreeNode(accountURI + ACC_SERVER_ID, basicAccountInfo.getServerId()));

        accountNodes.put(accountURI + ACC_SERVER_NONCE,
                         new TreeNode(accountURI + ACC_SERVER_NONCE,
                                      Base64.encode(basicAccountInfo.getServerNonce()),
                                      TreeNode.FORMAT_BINARY)
                         );

        accountNodes.put(accountURI + ACC_SERVER_PASSWORD,
                         new TreeNode(accountURI + ACC_SERVER_PASSWORD, basicAccountInfo.getServerPassword()));


        //Client
        accountNodes.put(accountURI + ACC_CLIENT_NONCE,
                         new TreeNode(accountURI + ACC_CLIENT_NONCE,
                                      Base64.encode(basicAccountInfo.getClientNonce()))
                         );

        accountNodes.put(accountURI + ACC_CLIENT_PASSWORD,
                         new TreeNode(accountURI + ACC_CLIENT_PASSWORD, basicAccountInfo.getClientPassword()));

        accountNodes.put(accountURI + ACC_CLIENT_USERNAME,
                         new TreeNode(accountURI + ACC_CLIENT_USERNAME, basicAccountInfo.getClientUsername()));


        return accountNodes;
    }

    /**
     * Returns the <code>SyncML</code> object corresponds to the bootstrap message for the given <code>SycnMLDM</code> object.
     * @param bootstrapNodes the Map of Tree Nodes
     *
     * @throws NotificationException if a error occurs during message creation
     * @return SyncML the <code>SyncML</code> corresponds to the bootstrap message
     */
    private SyncML getSyncMLBootstrapMessage(Map bootstrapNodes, BasicAccountInfo basicAccountInfo)
    throws NotificationException {

        SyncML   syncMLBootstrapMessage = null;
        SyncHdr  syncHdrMessage         = null;
        SyncBody syncBodyMessage        = null;

        //
        // SyncML Header Message
        //

        try {

            URL url = null;
            URL newUrlWithPort = null;

            try {
                url = new URL(basicAccountInfo.getServerAddress());
                String hostName = url.getHost();
                String protocol = url.getProtocol();
                String file = url.getFile();

                newUrlWithPort = new URL(protocol, hostName,
                                         basicAccountInfo.getServerPort(), file);
            } catch (MalformedURLException e) {
                throw new NotificationException("Error creating the serverURI", e);
            }


            syncHdrMessage = new SyncHdr(
                Constants.DTD_1_1,
                Constants.SYNCML_DM_1_1,
                new SessionID("0"),
                "0", // MsgID
                new Target(basicAccountInfo.getDeviceId()),
                new Source(newUrlWithPort.toExternalForm()),
                null, /* response URI */
                false, /* no reponse   */
                null, /* credentials  */
                null
                /* metadata     */
                );


            //
            // SyncML Body Message
            //
            Item[] items = Util.nodes2Items(bootstrapNodes, true, false, null);
            Add addCommand = new Add(
                    new CmdID(1),
                    false,
                    null ,
                    null ,
                    items);

            syncBodyMessage = new SyncBody(
                new AbstractCommand[] {addCommand},
                true
                /* final */
                );


            syncMLBootstrapMessage = new SyncML(syncHdrMessage, syncBodyMessage);


        } catch (RepresentationException ex) {
            throw new NotificationException("Error during creation of the SyncML message", ex);
        }

        return syncMLBootstrapMessage;
    }

}
