/**
 * Copyright (C) 2003-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */

package com.funambol.server.engine.dm.ejb;

import java.util.logging.Logger;

import javax.ejb.EJBException;
import javax.ejb.SessionContext;
import javax.naming.InitialContext;
import javax.rmi.PortableRemoteObject;

import com.funambol.framework.config.Configuration;
import com.funambol.framework.engine.dm.ManagementException;
import com.funambol.framework.logging.FunambolLogger;
import com.funambol.framework.logging.FunambolLoggerName;
import com.funambol.framework.notification.NotificationException;
import com.funambol.framework.notification.NotificationConstants;


import com.funambol.server.admin.ejb.AdminHomeRemote;
import com.funambol.server.admin.ejb.AdminRemote;
import com.funambol.server.engine.dm.ManagementEngine;
import com.funambol.framework.core.dm.bootstrap.BootStrap;

/**
 * Management ejb - Bean class
 *
 * @version $Id: ManagementBean.java,v 1.5 2007/01/11 19:31:01 nichele Exp $
 */
public class ManagementBean implements javax.ejb.SessionBean {

    // ------------------------------------------------------- Private constants

    private static final String ENV_ADMIN_NAME
        = "java:comp/env/ejb/AdminBean";

    // ------------------------------------------------------------ Private data
    private transient Logger log = null;
    private SessionContext sessionContext = null;


    // ------------------------------------------------------------- EJB methods

    public void ejbCreate() {
        log = FunambolLogger.getLogger(FunambolLoggerName.SERVER_DM_NOTIFICATION);
    }

    public void ejbRemove() {
    }

    public void ejbActivate() {
        log = FunambolLogger.getLogger(FunambolLoggerName.SERVER_DM_NOTIFICATION);
    }

    public void ejbPassivate() {
        log = null;
    }

    public void setSessionContext(SessionContext sessionContext) throws EJBException {
        this.sessionContext = sessionContext;
    }

    /**
     * Performs a bootstrap with the given bootstrap object
     *
     * @param bootstrap the bootstrap object with the neeed information to perfom
     *                  a bootstrap
     * @throws NotificationException if a error occurs
     */
    public void bootstrap(BootStrap bootstrap)
    throws ManagementException, NotificationException {
        Configuration conf = loadConfiguration();

        ManagementEngine mngtEngine = new ManagementEngine(conf);

        mngtEngine.bootstrap(bootstrap);
    }

    /**
     * Performs a bootstrap with the given bootstrap objects
     *
     * @param bootstraps the bootstrap array with the neeed information to perfom
     *                  the bootstraps
     * @throws NotificationException if a error occurs
     */
    public void bootstrap(BootStrap[] bootstraps)
    throws ManagementException, NotificationException {
        Configuration conf = loadConfiguration();

        ManagementEngine mngtEngine = new ManagementEngine(conf);

        mngtEngine.bootstrap(bootstraps);
    }

    /**
     * Sends a notification message to the device with the given phoneNumber
     * @param messageType the type of the notification message as define in <code>NotificationConstants<code>
     * @param transportType the type of the transport as define in <code>NotificationConstants<code>
     * @param phoneNumber the phone number
     * @param operation application specific operation to be performed
     * @param info application specific detail information
     *
     * @throws ManagementException if a error occurs in the management engine,
     *         NotificationException if an error occurs in the notification process
     */
    public void sendNotification( int    messageType  ,
                                  int    transportType,
                                  String phoneNumber  ,
                                  String operation    ,
                                  String info         )
    throws ManagementException, NotificationException {

        Configuration conf = loadConfiguration();

        ManagementEngine mngtEngine = new ManagementEngine(conf);

        mngtEngine.sendNotification(
            messageType,
            transportType,
            phoneNumber,
            operation,
            info
        );
    }

    /**
     * Executes the management sequence identified by the given operation name.
     *
     * @param phoneNumber the phone number
     * @param operation the management operation name
     * @param info application specific detail information
     *
     * @throws ManagementException if a error occurs in the management engine,
     *         NotificationException if an error occurs in the notification process
     */
    public void executeManagementOperation( String phoneNumber  ,
                                            String operation    ,
                                            String info         )
    throws ManagementException, NotificationException {

        sendNotification(
            NotificationConstants.NOTIFICATION_MESSAGE_TYPE_GENERIC,
            NotificationConstants.TRANSPORT_TYPE_WAP,
            phoneNumber,
            operation,
            info
        );
    }



    // --------------------------------------------------------- Private methods

    /**
     * Load the configuration through the management bean
     * @throws ManagementException if a error occurs
     * @return Configuration
     */
    private Configuration loadConfiguration()
    throws ManagementException {

        Configuration c = null;
        try {

            InitialContext ctx = new InitialContext();

            Object obj = ctx.lookup(ENV_ADMIN_NAME);
            AdminHomeRemote h = (AdminHomeRemote)PortableRemoteObject.narrow(obj,
                AdminHomeRemote.class);

            AdminRemote m = h.create();

            c = m.getConfig();

            m.remove();

        } catch (Exception e) {
            log.throwing("ManagementBean", "loadConfiguration", e);
            throw new ManagementException("Error reading the configuration", e);
        }
        return c;
    }



}
