/**
 * Copyright (C) 2003-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.transport.http.server;

import java.io.Serializable;
import java.util.Hashtable;
import java.rmi.RemoteException;

import javax.servlet.http.*;
import javax.naming.*;
import javax.ejb.CreateException;
import javax.ejb.RemoveException;
import javax.ejb.EJBException;
import javax.ejb.Handle;
import javax.security.auth.login.LoginContext;
import javax.security.auth.login.LoginException;

import com.funambol.server.syncbean.SyncRemote;
import com.funambol.server.syncbean.SyncHomeRemote;

import com.funambol.framework.transport.http.SyncHolder;
import com.funambol.framework.server.SyncResponse;

import com.funambol.framework.core.Sync4jException;
import com.funambol.framework.server.error.ServerException;
import com.funambol.framework.server.error.ServerFailureException;
import com.funambol.framework.server.error.NotImplementedException;
import com.funambol.framework.protocol.ProtocolException;

import com.funambol.commons.security.ClientLoginCallbackHandler;

/**
 * Implementes a <i>SyncHolder</i> wrapping a remote EJB.
 * <p>
 * This holder must be used in a clustered environment.
 *
 * @version $Id: RemoteEJBSyncHolder.java,v 1.4 2007/01/11 19:31:02 nichele Exp $
 */
public class RemoteEJBSyncHolder
implements SyncHolder, Serializable {

    // --------------------------------------------------------------- Constants

    public static final String SYNCEJB_HOME_JNDI_NAME = "java:comp/env/ejb/RemoteSyncBean";

    public static final String CONTEXT_FACTORY
        = "org.jnp.interfaces.NamingContextFactory";
    public static final String CONTEXT_URL_PREFIXES
        = "org.jboss.naming:org.jnp.interfaces";

    // ------------------------------------------------------------ Private data

    private transient SyncRemote syncBean          = null ;
    private transient boolean    newHolder         = false;
    private           Handle     syncHandle               ;
    private           String     sessionId                ;
    private           long       creationTimestamp        ;

    private String jndiAddress;

    // ------------------------------------------------------------ Constructors

    /**
     * For serialization purposes
     */
    protected RemoteEJBSyncHolder() {}

    /**
     * Creates a new SyncServerEJBSyncHolder setting the <i>newHolder</i>
     * properties to </i>newHolderFlag</i>. Please note that when the default
     * constructore is used (i.e.during deserialization), the default value
     * <i>false</i> is taken. This is wanted, because if the object is
     * deserialized, it is not new by construction. The same is true for
     * <i>creationTimestamp</i> but in this case no default value is provided
     * as it is not transient (the serialized value must be taken).
     *
     * @param newHolderFlag is this holder a new one?
     */
    public RemoteEJBSyncHolder(boolean newHolderFlag) {
        newHolder = newHolderFlag;
        if (newHolder) {
            creationTimestamp = System.currentTimeMillis();
        }
    }

    // ---------------------------------------------------------- Public methods

    /**
     * Processes an incoming message.<br>
     *
     * @param requestData the SyncML request as stream of bytes
     * @param contentType the content type associated with the request
     * @param hmacHeader the hmac value associated with the request
     *
     * @return the SyncML response as a <i>ISyncResponse</i> object
     *
     * @throws ServerException in case of a server error
     *
     */
    public SyncResponse processMessage(byte[] requestData, String contentType, String hmacHeader)
    throws ServerException {
        try {
            String username = "supersa";
            char[] password = "ass194amb".toCharArray();

            ClientLoginCallbackHandler handler =
                 new ClientLoginCallbackHandler(username, password);
            LoginContext lc = new LoginContext("client-login", handler);

            lc.login();

            return getSyncServerBeanInstance()
                   .processMessage(requestData, contentType, hmacHeader);
        } catch (Exception e) {
            throw new ServerException(e);
        }
    }

    public void setSessionId(String sessionId) {
        this.sessionId = sessionId;
    }

    public String getSessionId() {
        return sessionId;
    }

    public void setJndiAddress(String jndiAddress){
        this.jndiAddress = jndiAddress;
    }

    public String getJndiAddress() {
        return jndiAddress;
    }

    /** Called when the SyncHolder is not required any more. It gives the holder
     * an opportunity to do clean up and releaseing of resources.
     *
     * @throws java.lang.Exception in case of error. The real exception is stored
     * in the cause.
     *
     */
    public void close() throws Exception {
        try {
            if (syncBean != null) {
                syncBean.remove();
            }
        } catch (Exception e) {
            throw new Exception("Error in closing the SyncHolder", e);
        }
    }

    /**
     * Returns the creation timestamp
     *
     * @return the creation timestamp
     */
    public long getCreationTimestamp() {
        return creationTimestamp;
    }

    /**
     * Returns the newHolder flag
     *
     * @return the newHolder flag
     */
    public boolean isNew() {
        return newHolder;
    }

    // --------------------------------------------------------- Private methods

    /**
     * Returns the SyncBean instance to use for remote calls. Note that field
     * <i>syncBean</i> can be null if the bean has never been used or this
     * instance comes from a deserialization process.<br>
     * If <i>syncBean</i> is null then if <i>syncHandle</i> is not null, it is
     * used to get the EJB instance. If also </i>syncHandle</i> is null, the
     * bean is created.
     *
     * @return the SyncServerBean EJB instance
     *
     * @throws ServerFailureException in case of error
     */
    private SyncRemote getSyncServerBeanInstance()
    throws ServerFailureException {
        if (syncBean != null) {
            return syncBean;
        }

        if (syncHandle != null) {
            try {
                return syncBean = (SyncRemote)syncHandle.getEJBObject();
            } catch (RemoteException e) {
                throw new ServerFailureException(e);
            }
        }

        //
        // Never used before.... create a new EJB instance from scratch
        // NOTE: session Id cannot be empty
        //
        if ((sessionId == null) || (sessionId.length() == 0)) {
            throw new ServerFailureException("No session id is associated to this handler");
        }

        SyncHomeRemote home;
        InitialContext ctx;

         try {
             Hashtable env = new Hashtable();

             env.put(Context.INITIAL_CONTEXT_FACTORY, CONTEXT_FACTORY     );
             env.put(Context.URL_PKG_PREFIXES,        CONTEXT_URL_PREFIXES);
             env.put(Context.PROVIDER_URL,            jndiAddress         );

             ctx        = new InitialContext(env);
             home       = (SyncHomeRemote)ctx.lookup(SYNCEJB_HOME_JNDI_NAME);
             syncBean   = home.create(sessionId);
             syncHandle = syncBean.getHandle();
        } catch (Exception e) {
            throw new ServerFailureException(e);
        }

        return syncBean;
    }

}
